<?php

namespace Drupal\static_build\Controller;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Routing\LocalRedirectResponse;
use Drupal\Core\Session\AccountProxyInterface;
use Drupal\Core\Url;
use Drupal\static_build\Plugin\StaticBuilderHelperInterface;
use Drupal\static_build\Plugin\StaticBuilderPluginInterface;
use Drupal\static_build\Plugin\StaticBuilderPluginManagerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\RequestStack;
use Throwable;

/**
 * Provides various controllers to interact with releases on build context.
 */
class ReleaseController extends ControllerBase {

  /**
   * Request stack.
   *
   * @var \Symfony\Component\HttpFoundation\Request|null
   */
  protected $request;

  /**
   * The static builder plugin manager.
   *
   * @var \Drupal\static_build\Plugin\StaticBuilderPluginManagerInterface
   */
  protected $staticBuilderPluginManager;

  /**
   * The static builder helper.
   *
   * @var \Drupal\static_build\Plugin\StaticBuilderHelperInterface
   */
  protected $staticBuilderHelper;

  /**
   * The static builder plugin.
   *
   * @var \Drupal\static_build\Plugin\StaticBuilderPluginInterface
   */
  protected $builder;

  /**
   * Static Builder Releases Controller constructor.
   *
   * @param \Symfony\Component\HttpFoundation\RequestStack $requestStack
   *   The request stack.
   * @param \Drupal\Core\Session\AccountProxyInterface $currentUser
   *   The current user.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   Config factory.
   * @param \Drupal\static_build\Plugin\StaticBuilderPluginManagerInterface $staticBuilderPluginManager
   *   The static builder plugin manager.
   * @param \Drupal\static_build\Plugin\StaticBuilderHelperInterface $staticBuilderHelper
   *   The static builder helper.
   */
  public function __construct(
    RequestStack $requestStack,
    AccountProxyInterface $currentUser,
    ConfigFactoryInterface $config_factory,
    StaticBuilderPluginManagerInterface $staticBuilderPluginManager,
    StaticBuilderHelperInterface $staticBuilderHelper
  ) {
    $this->request = $requestStack->getCurrentRequest();
    $this->currentUser = $currentUser;
    $this->configFactory = $config_factory;
    $this->staticBuilderPluginManager = $staticBuilderPluginManager;
    $this->staticBuilderHelper = $staticBuilderHelper;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('request_stack'),
      $container->get('current_user'),
      $container->get('config.factory'),
      $container->get('plugin.manager.static_builder'),
      $container->get('static_build.static_builder_helper')
    );
  }

  /**
   * Shows a list of releases.
   *
   * @param string $builderId
   *   Builder id.
   * @param string $runMode
   *   A run mode to get releases for. Usually live or preview.
   *
   * @return array
   *   A render array as expected by the renderer.
   */
  public function listReleases(string $builderId, string $runMode): array {
    $builder = NULL;
    try {
      $builder = $this->staticBuilderPluginManager->getInstance([
        'plugin_id' => $builderId,
        'configuration' => ['run-mode' => $runMode],
      ]);
      $releaseManager = $builder->getReleaseManager();
    }
    catch (Throwable $e) {
      return $list['static_build_releases_list'] = [
        '#markup' => '<p>' . $this->t('Unable to list releases:') . ' ' . $e->getMessage() . '</p>',
      ];
    }

    $allReleases = $releaseManager->getAllReleases();

    $rows = [];
    $ok = json_decode('"\u2714\ufe0f"', FALSE);
    $ko = json_decode('"\u274c"', FALSE);
    foreach ($allReleases as $delta => $release) {
      $releaseTask = $release->task($builder->getTaskId());
      $row = [];
      $row['id'] = $release->uniqueId();

      // The progress bar should only appear once, for the latest release. This
      // way, we avoid problems with stale releases, that make this page reload
      // again an again in an infinite loop until the release is marked as
      // failed.
      if ($delta === 0 && !$releaseTask->isDone() && !$releaseTask->isFailed()) {
        $row['speed'] = '--';
        $row['percentage']['data'] = [
          '#theme' => 'progress_bar',
          '#attached' => [
            'drupalSettings' => [
              'batch' => [
                'delay' => 3000,
                'errorMessage' => $this->t('An error has occurred, please reload this page.'),
                'initMessage' => $this->t('Gathering build data...'),
                'uri' => Url::fromRoute('static_build.releases_list.running_batch_data', [
                  'builderId' => $builderId,
                  'runMode' => $runMode,
                  'destination' => $this->request->getRequestUri(),
                ])->toString(),
              ],
            ],
            'library' => [
              'static_suite/batch',
            ],
          ],
        ];
      }
      else {
        $row['speed'] = $releaseTask->getProcessBenchmark() . ' ' . $this->t('seconds');
        $row['percentage'] = $releaseManager->getTaskSupervisor()
            ->getTaskPercentage($release->uniqueId(), $builder->getTaskId()) . "%";
      }

      $row['current'] = $releaseManager->isCurrent($release->uniqueId()) ? $ok : $ko;
      $row['build-done'] = $releaseTask->isDone() ? $ok : $ko;
      $row['build-log']['data'] = [
        '#type' => 'inline_template',
        '#template' => '<a href="' . Url::fromRoute('static_build.log_viewer', [
            'builderId' => $builderId,
            'runMode' => $runMode,
            'uniqueId' => $release->uniqueId(),
          ])->toString() . '" target="_blank">' . $this->t('View') . '</a>',
      ];
      $rows[] = $row;
    }

    $headers = [
      $this->t('Unique ID'),
      $this->t('Build speed (~@medianSpeed secs)', [
        '@medianSpeed' => $releaseManager->getTaskSupervisor()
          ->getMedianTaskTime($builder->getTaskId()),
      ]),
      $this->t('Build percentage'),
      $this->t('Is current'),
      $this->t('Successfully built'),
      $this->t('Build log'),
    ];

    $headerMessage = $this->t(
      'Releases built by <strong>"@builder"</strong> on <strong>"@run-mode"</strong> mode.',
      [
        '@builder' => $builder->label(),
        '@run-mode' => $runMode,
        '@cloud' => StaticBuilderPluginInterface::HOST_MODE_CLOUD,
      ]
    );

    $subHeaderMessage = NULL;
    if ($runMode === StaticBuilderPluginInterface::RUN_MODE_PREVIEW) {
      $subHeaderMessage = $this->t(
        'Keep in mind that "@cloud" builders do not support "@run-mode" mode.',
        [
          '@cloud' => StaticBuilderPluginInterface::HOST_MODE_CLOUD,
          '@run-mode' => $runMode,
        ]
      );
    }

    $list['releases_list_header'] = [
      '#markup' => '<p>' . $headerMessage . ' ' . $subHeaderMessage . '</p>',
    ];

    if ($this->currentUser->hasPermission('run builds on demand')) {
      $list['run_new_build'] = [
        '#theme' => 'menu_local_action',
        '#link' => [
          'title' => $this->t('Run new build'),
          'url' => Url::fromRoute('static_build.run_build_on_demand', [
            'builderId' => $builder->getPluginId(),
            'runMode' => $runMode,
            'destination' => $this->request->getRequestUri(),
          ]),
        ],
        // Without this workaround, the action links will be rendered as <li>
        // with no wrapping <ul> element.
        // @todo Find a better approach to wrap local action links in a <ul>
        //   (https://www.drupal.org/node/3181052) or find a way for derivers to
        //   work with "appears_on" keys that have parameters on them.
        '#prefix' => '<ul class="action-links">',
        '#suffix' => '</ul>',
      ];
    }

    $list['releases_list'] = [
      '#type' => 'table',
      '#header' => $headers,
      '#rows' => $rows,
      '#empty' => $this->t('No releases available. Please edit some content or run a build manually.'),
      '#attributes' => [
        'class' => ['block-add-table'],
      ],
    ];

    return $list;
  }

  /**
   * Get data from the release currently being built.
   *
   * Gives an estimated percentage.
   *
   * @param string $builderId
   *   Builder id.
   * @param string $runMode
   *   A run mode. Usually live or preview.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   JSON response.
   */
  public function getRunningReleaseData(string $builderId, string $runMode): JsonResponse {
    return new JsonResponse(
      $this->staticBuilderHelper->getRunningBuildData($builderId, $runMode)
    );
  }

  /**
   * Get batch data from the release currently being built.
   *
   * @param string $builderId
   *   Builder id.
   * @param string $runMode
   *   A run mode. Usually live or preview.
   */
  public function getRunningReleaseBatchData(string $builderId, string $runMode) {
    // Redirect when finished.
    if ($this->request->query->get('destination') && $this->request->query->get('op') === 'finished') {
      return new LocalRedirectResponse($this->request->query->get('destination'));
    }

    $data = $this->staticBuilderHelper->getRunningBuildData($builderId, $runMode);

    // Set 100 if the build process is already finished.
    $percentage = (!$data['unique-id'] && $data['percentage'] === 0) ? 100 : $data['percentage'];

    return new JsonResponse([
      'status' => TRUE,
      'percentage' => (string) $percentage,
      'message' => $percentage === 100 ? $this->t('Please wait') : $this->t('@seconds seconds...', ['@seconds' => $data['elapsed-seconds']]),
      'label' => $percentage === 100 ? $this->t('Finished') : $this->t('Building...'),
    ]);
  }

}
