<?php

namespace Drupal\static_deploy\Controller;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Language\LanguageManagerInterface;
use Drupal\Core\Routing\LocalRedirectResponse;
use Drupal\Core\Session\AccountProxyInterface;
use Drupal\Core\Url;
use Drupal\static_build\Plugin\StaticBuilderPluginInterface;
use Drupal\static_build\Plugin\StaticBuilderPluginManagerInterface;
use Drupal\static_deploy\Plugin\StaticDeployerHelperInterface;
use Drupal\static_deploy\Plugin\StaticDeployerPluginManagerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\RequestStack;
use Throwable;

/**
 * Provides various controllers to interact with releases on deploy context.
 */
class ReleaseController extends ControllerBase {

  /**
   * The request stack.
   *
   * @var \Symfony\Component\HttpFoundation\Request|null
   */
  protected $request;

  /**
   * Drupal Language Manager.
   *
   * @var \Drupal\Core\Language\LanguageManagerInterface
   */
  protected $languageManager;

  /**
   * The static builder plugin manager.
   *
   * @var \Drupal\static_build\Plugin\StaticBuilderPluginManagerInterface
   */
  protected $staticBuilderPluginManager;

  /**
   * The static deployer plugin manager.
   *
   * @var \Drupal\static_deploy\Plugin\StaticDeployerPluginManagerInterface
   */
  protected $staticDeployerPluginManager;

  /**
   * The static deployer helper.
   *
   * @var \Drupal\static_deploy\Plugin\StaticDeployerHelperInterface
   */
  protected $staticDeployerHelper;

  /**
   * StaticBuilderReleasesController constructor.
   *
   * @param \Symfony\Component\HttpFoundation\RequestStack $requestStack
   *   The request stack.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   Config factory.
   * @param \Drupal\Core\Language\LanguageManagerInterface $languageManager
   *   Drupal language manager.
   * @param \Drupal\static_build\Plugin\StaticBuilderPluginManagerInterface $staticBuilderPluginManager
   *   The static builder plugin manager.
   * @param \Drupal\static_deploy\Plugin\StaticDeployerPluginManagerInterface $staticDeployerPluginManager
   *   The static deployer plugin manager.
   * @param \Drupal\static_deploy\Plugin\StaticDeployerHelperInterface $staticDeployerHelper
   *   The static deployer helper.
   */
  public function __construct(
    RequestStack $requestStack,
    AccountProxyInterface $currentUser,
    ConfigFactoryInterface $config_factory,
    LanguageManagerInterface $languageManager,
    StaticBuilderPluginManagerInterface $staticBuilderPluginManager,
    StaticDeployerPluginManagerInterface $staticDeployerPluginManager,
    StaticDeployerHelperInterface $staticDeployerHelper
  ) {
    $this->request = $requestStack->getCurrentRequest();
    $this->currentUser = $currentUser;
    $this->configFactory = $config_factory;
    $this->languageManager = $languageManager;
    $this->staticBuilderPluginManager = $staticBuilderPluginManager;
    $this->staticDeployerPluginManager = $staticDeployerPluginManager;
    $this->staticDeployerHelper = $staticDeployerHelper;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('request_stack'),
      $container->get('current_user'),
      $container->get('config.factory'),
      $container->get('language_manager'),
      $container->get('plugin.manager.static_builder'),
      $container->get('plugin.manager.static_deployer'),
      $container->get('static_deploy.static_deployer_helper'),
    );
  }

  /**
   * Shows a list of LIVE releases with deploy info.
   *
   * @param string $deployerId
   *   Static Deployer Id.
   * @param string $builderId
   *   Static Builder Id.
   *
   * @return array
   *   A render array as expected by the renderer.
   */
  public function listReleases(string $deployerId, string $builderId): array {
    try {
      $builder = $this->staticBuilderPluginManager->getInstance([
        'plugin_id' => $builderId,
        'configuration' => ['run-mode' => StaticBuilderPluginInterface::RUN_MODE_LIVE],
      ]);
      $deployer = $this->staticDeployerPluginManager->getInstance([
        'plugin_id' => $deployerId,
        'configuration' => ['builder-id' => $builderId],
      ]);
      $releaseManager = $deployer->getReleaseManager();
    }
    catch (Throwable $e) {
      return $list['static_deploy_releases_list'] = [
        '#markup' => '<p>' . $this->t('Unable to list releases:') . ' ' . $e->getMessage() . '</p>',
      ];
    }

    $allReleases = $releaseManager->getAllReleases();

    $rows = [];
    $ok = json_decode('"\u2714\ufe0f"', FALSE);
    $ko = json_decode('"\u274c"', FALSE);
    foreach ($allReleases as $delta => $release) {
      $releaseTask = $release->task($deployer->getTaskId());
      $row = [];
      $row['id'] = $release->uniqueId();

      // The progress bar should only appear once, for the latest release. This
      // way, we avoid problems with stale releases, that make this page reload
      // again an again in an infinite loop until the release is detected as
      // failed.
      if ($delta === 0 && !$releaseTask->isDone() && !$releaseTask->isFailed()) {
        $row['speed'] = '--';
        $row['percentage']['data'] = [
          '#theme' => 'progress_bar',
          '#attached' => [
            'drupalSettings' => [
              'batch' => [
                'delay' => 3000,
                'errorMessage' => $this->t('An error has occurred, please reload this page.'),
                'initMessage' => $this->t('Gathering deploy data...'),
                'uri' => Url::fromRoute('static_deploy.releases_list.running_batch_data', [
                  'deployerId' => $deployerId,
                  'builderId' => $builderId,
                  'destination' => $this->request->getRequestUri(),
                ])->toString(),
              ],
            ],
            'library' => [
              'static_suite/batch',
            ],
          ],
        ];
      }
      else {
        $row['speed'] = $releaseTask->getProcessBenchmark() . ' ' . $this->t('seconds');
        $row['percentage'] = $releaseManager->getTaskSupervisor()
            ->getTaskPercentage($release->uniqueId(), $deployer->getTaskId()) . '%';
      }

      $row['deploy-started'] = $releaseTask->isStarted() ? $ok : $ko;
      $row['deploy-done'] = $releaseTask->isDone() ? $ok : $ko;
      $row['deploy-log']['data'] = [
        '#type' => 'inline_template',
        '#template' => '<a href="' . Url::fromRoute(
            'static_deploy.log_viewer', [
              'builderId' => $builderId,
              'deployerId' => $deployerId,
              'uniqueId' => $release->uniqueId(),
            ]
          )->toString() . '" target="_blank">' . $this->t('View') . '</a>',
      ];
      $rows[] = $row;
    }

    $headers = [
      $this->t('Unique ID'),
      $this->t('Deploy speed (~@medianSpeed secs)', [
        '@medianSpeed' => $releaseManager->getTaskSupervisor()
          ->getMedianTaskTime($deployer->getTaskId()),
      ]),
      $this->t('Deploy percentage'),
      $this->t('Deploy started'),
      $this->t('Deploy done'),
      $this->t('Deploy log'),
    ];

    $list['releases_list_header'] = [
      '#markup' => '<p>' .
        $this->t(
          'Releases deployed by <strong>"@deployer"</strong> (built by <strong>"@builder"</strong> on <strong>"@run-mode"</strong> mode). Keep in mind that releases created by "@cloud" builders are automatically deployed by the CI/CD service in charge of the build process.',
          [
            '@deployer' => $deployer->label(),
            '@builder' => $builder->label(),
            '@run-mode' => StaticBuilderPluginInterface::RUN_MODE_LIVE,
            '@cloud' => StaticBuilderPluginInterface::HOST_MODE_CLOUD,
          ])
        . '</p>',
    ];

    if ($this->currentUser->hasPermission('run deploys on demand')) {
      $list['run_new_build'] = [
        '#theme' => 'menu_local_action',
        '#link' => [
          'title' => $this->t('Run new deploy'),
          'url' => Url::fromRoute('static_deploy.run_deploy_on_demand', [
            'deployerId' => $deployerId,
            'builderId' => $builderId,
            'destination' => $this->request->getRequestUri(),
          ]),
        ],
        // Without this workaround, the action links will be rendered as <li>
        // with no wrapping <ul> element.
        // @todo Find a better approach to wrap local action links in a <ul>
        //   (https://www.drupal.org/node/3181052) or find a way for derivers to
        //   work with "appears_on" keys that have parameters on them.
        '#prefix' => '<ul class="action-links">',
        '#suffix' => '</ul>',
      ];
    }

    $list['releases_list'] = [
      '#type' => 'table',
      '#header' => $headers,
      '#rows' => $rows,
      '#empty' => $this->t('No releases available. Please edit some content or run a build manually.'),
      '#attributes' => [
        'class' => ['block-add-table'],
      ],
    ];

    return $list;
  }

  /**
   * Get data from the release currently being deployed.
   *
   * Gives an estimated percentage.
   *
   * @param string $deployerId
   *   Static Deployer Id.
   * @param string $builderId
   *   Static Builder Id.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   JSON response.
   */
  public function getRunningReleaseData(string $deployerId, string $builderId): JsonResponse {
    return new JsonResponse(
      $this->staticDeployerHelper->getRunningDeployData($deployerId, $builderId)
    );
  }

  /**
   * Get batch data from the release currently being deployed.
   *
   * @param string $deployerId
   *   Static Deployer Id.
   * @param string $builderId
   *   Static Builder Id.
   */
  public function getRunningReleaseBatchData(string $deployerId, string $builderId) {
    // Redirect when finished.
    if ($this->request->query->get('destination') && $this->request->query->get('op') === 'finished') {
      return new LocalRedirectResponse($this->request->query->get('destination'));
    }

    $data = $this->staticDeployerHelper->getRunningDeployData($deployerId, $builderId);

    // Set 100 if the deploy process is already finished.
    $percentage = (!$data['unique-id'] && $data['percentage'] === 0) ? 100 : $data['percentage'];

    return new JsonResponse([
      'status' => TRUE,
      'percentage' => (string) $percentage,
      'message' => $percentage === 100 ? $this->t('Please wait') : $this->t('@seconds seconds...', ['@seconds' => $data['elapsed-seconds']]),
      'label' => $percentage === 100 ? $this->t('Finished') : $this->t('Deploying...'),
    ]);
  }

}
